import torch
from torchvision import models, transforms
import torch.nn as nn
import torch.nn.functional as F

def init_weights(m):
    if isinstance(m, nn.Linear):
        nn.init.xavier_uniform_(m.weight)

def build_model(model_name, num_classes, use_pretrained=False):
    model_ft = None
    if model_name == "resnet":
        """ Resnet18
        """
        model_ft = models.resnet18(pretrained=use_pretrained)
        num_ftrs = model_ft.fc.in_features
        model_ft.fc = nn.Linear(num_ftrs, num_classes)
        # input_size = 224
    
    elif model_name == "CNN":
        model_ft = Net(num_classes)
        model_ft.apply(init_weights)
    
    elif model_name == "MLP":
        model_ft = MLP(num_classes)
        model_ft.apply(init_weights)

    elif model_name == "MLP1":
        model_ft = MLP1(num_classes)
        model_ft.apply(init_weights)
    elif model_name == "MLP2":
        model_ft = MLP2(num_classes)
        model_ft.apply(init_weights)
    elif model_name == "MLP3":
        model_ft = MLP3(num_classes)
        model_ft.apply(init_weights)
    elif model_name == "MLP4":
        model_ft = MLP4(num_classes)
        model_ft.apply(init_weights)
    elif model_name == "MLP5":
        model_ft = MLP5(num_classes)
        model_ft.apply(init_weights)
    else:
        print("Invalid model name, exiting...")
        exit()
    
    return model_ft


# class MLP(nn.Module):
#     def __init__(self, num_classes):
#         super().__init__()
#         self.fc1 = nn.Linear(784, 32)
#         self.fc2 = nn.Linear(32, 10)
#         self.dropout = nn.Dropout(p=0.2)

#     def forward(self, x):
#         x = F.relu(self.fc1(x.reshape(-1, 784)))
#         x = self.dropout(x)
#         x = F.relu(self.fc2(x))
#         return x
class MLP(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.fc1 = nn.Linear(784, 10)

    def forward(self, x):
        x = F.relu(self.fc1(x.reshape(-1, 784)))
        return x


class Net(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, num_classes)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1) # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

    
class Linear(nn.Module):
    def __init__(self, dataset):
        super().__init__()
        if dataset == "MNIST":
            self.fc = nn.Linear(784, 1)
        else:
            self.fc = nn.Linear(3072, 1)

    def forward(self, x):
        x = torch.sigmoid(self.fc(x))
        return x

class MLP1(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.fc1 = nn.Linear(784, 10)
    def forward(self, x):
        x = F.relu(self.fc1(x.reshape(-1, 784)))
        return x


class MLP2(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.fc1 = nn.Linear(784, 1024)
        self.fc2 = nn.Linear(1024, 10)
        self.dropout = nn.Dropout(p=0.2)

    def forward(self, x):
        x = F.relu(self.fc1(x.reshape(-1, 784)))
        x = self.dropout(x)
        x = F.relu(self.fc2(x))
        return x


class MLP3(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.fc1 = nn.Linear(784, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, 10)
        self.dropout = nn.Dropout(p=0.2)

    def forward(self, x):
        x = F.relu(self.fc1(x.reshape(-1, 784)))
        x = self.dropout(x)
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        return x


class MLP4(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.fc1 = nn.Linear(784, 1024)
        self.fc2 = nn.Linear(1024, 256)
        self.fc3 = nn.Linear(256, 10)
        self.dropout = nn.Dropout(p=0.2)

    def forward(self, x):
        x = F.relu(self.fc1(x.reshape(-1, 784)))
        x = self.dropout(x)
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        return x


class MLP5(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.fc1 = nn.Linear(784, 1024)
        self.fc2 = nn.Linear(1024, 512)
        self.fc3 = nn.Linear(512, 256)
        self.fc4 = nn.Linear(256, 10)
        self.dropout = nn.Dropout(p=0.2)

    def forward(self, x):
        x = F.relu(self.fc1(x.reshape(-1, 784)))
        x = self.dropout(x)
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        x = F.relu(self.fc4(x))
        return x